@@ -1,9 +1,16 @@
 from cannerflow.utils import *
-
+from io import BytesIO
+import requests
+import pandas as pd
 __all__ = ["Query"]
 
 
 class Query(object):
+    @staticmethod
+    def __get_pd_by_parquet_file(url):
+        r = requests.get(url)
+        return pd.read_parquet(BytesIO(r.content))
+
     def __init__(
         self,
         workspace_id,
@@ -11,7 +18,8 @@ class Query(object):
         sql,
         cache_refresh,
         cache_ttl,
-        data_format
+        data_format,
+        fetch_by='storage',
     ):
         self.workspace_id = workspace_id
         self.request = request
@@ -25,6 +33,7 @@ class Query(object):
             cache_refresh=cache_refresh,
             cache_ttl=cache_ttl,
             sql=sql)
+        self.fetch_by = fetch_by
         self.data = []
         self.row_count = 0
         self.id = None
@@ -45,7 +54,7 @@ class Query(object):
         index = 0
         page = 1
         page_size = 50000
-        self.get_first() # initial columns and row count
+        self.get_first()  # initial columns and row count
 
         # fetch first page
         result = self.request.post(gen_sql_result_payload(self.id, page_size, 0)).get(
@@ -53,20 +62,19 @@ class Query(object):
         data = result['result']
 
         while (page - 1) * page_size + index < self.row_count:
-            if index == page_size:
+            if index == len(data):
                 # fetch new page
                 result = self.request.post(gen_sql_result_payload(self.id, page_size, page_size * page)).get('sqlResultPagination')
                 data = result['result']
                 index = 0
                 page += 1
-            if len(data) > index:
+            elif index < len(data):
                 rtn_data = data_factory(data_format=self.data_format, columns=self.columns, data=[data[index]])
                 yield rtn_data
                 index += 1
             else:
                 break
 
-
     @property
     def columns(self):
         self.__ensure_columns()
@@ -91,6 +99,27 @@ class Query(object):
         self.get_sql_query()
 
     def get_sql_result(self, limit, offset):
+        if self.fetch_by is 'storage':
+            self.__get_sql_result_by_storage(limit, offset)
+        else:
+            self.__get_sql_result_by_restful(limit, offset)
+
+    def __get_sql_result_by_storage(self, limit, offset):
+        limit = limit or self.limit
+        offset = offset or self.offset
+        query_id = self.id
+        workspace_id = self.workspace_id
+        try:
+            urls = self.request.get(f'api/v1/query/{query_id}/result/signedUrls?workspaceId={workspace_id}').get('signedUrls')
+            dfs = [self.__get_pd_by_parquet_file(url) for url in urls]
+            df = pd.concat(dfs)
+        except:
+            df = pd.DataFrame({})
+        self._columns = list(map(lambda column: {'name': column}, list(df.columns)))
+        self.row_count = df.shape[0]
+        self.data = df.iloc[offset:limit + offset, :].to_numpy().tolist()
+
+    def __get_sql_result_by_restful(self, limit, offset):
         limit = limit or self.limit
         offset = offset or self.offset
         result = self.request.post(gen_sql_result_payload(self.id, limit, offset)).get('sqlResultPagination')
